package com.winhc.dataworks.flow.touch;

import cn.hutool.cache.CacheUtil;
import cn.hutool.cache.impl.TimedCache;
import com.aliyuncs.dataworks_public.model.v20180601.CreateManualDagResponse;
import com.helospark.lightdi.LightDi;
import com.helospark.lightdi.LightDiContext;
import com.helospark.lightdi.annotation.Autowired;
import com.helospark.lightdi.annotation.Service;
import com.winhc.dataworks.flow.touch.bean.*;
import com.winhc.dataworks.flow.touch.configuration.SchemaInit;
import com.winhc.dataworks.flow.touch.service.OdpsService;
import com.winhc.dataworks.flow.touch.service.TouchService;
import com.winhc.dataworks.flow.touch.utils.DateUtils;
import com.winhc.dataworks.flow.touch.utils.DingUtils;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.cli.*;

import java.io.FileNotFoundException;
import java.time.Duration;
import java.time.LocalDateTime;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @Author: XuJiakai
 * @Date: 2020/6/22 10:45
 * @Description:
 */
@Slf4j
@Service
public class Main {
    @Autowired
    private TouchService touchService;
    private static final Options options = new Options();


    @Autowired
    private DingUtils dingUtils;

    static {
        options.addOption("q", "query", false, "选填，查询昨天company增量表是否有数据");
        options.addOption("d", "debug", false, "选填，启用debug模式，只输出任务参数不进行调度");
        options.addOption("s", "singleJob", false, "选填，是否单独触发一个业务流程");
        options.addOption("b", "bizDate", true, "选填，业务时间[2020-07-07],默认为昨天");
        options.addOption("f", "fileName", true, "必填，yaml文件");

        options.addOption("flow", "flow", true, "单任务必填，业务流程名");
        options.addOption("task", "taskName", true, "单任务必填，任务名");
    }

    private static void verify(CommandLine commandLine) {
        try {
            if (!commandLine.hasOption("f")) {
                throw new RuntimeException();
            }
            if (commandLine.hasOption("s")) {
                if (!commandLine.hasOption("flow") || !commandLine.hasOption("task")) {
                    throw new RuntimeException();
                }
            }
        } catch (Exception e) {
            HelpFormatter formatter = new HelpFormatter();
            formatter.printHelp("很棒，参数错误", options);
            System.exit(-1);
        }
    }


    public static void main(String[] args) {
        LightDiContext context = LightDi.initContextByPackage(Main.class.getPackage().getName());
        DingUtils dd = context.getBean(DingUtils.class);
        CommandLineParser parser = new DefaultParser();
        try {
            CommandLine commandLine = parser.parse(options, args);
            if (commandLine.hasOption("q")) {
                OdpsService bean = context.getBean(OdpsService.class);
                String ds = DateUtils.getYesterday().replace("-", "");
                List<String> emptyTable = bean.getEmptyTable(ds);
                HashSet<String> set = new HashSet<>(emptyTable);
                if (set.contains("company")) {
                    log.error("company表{}分区数据为空，调度程序中止!", ds);
                    dd.send("company表" + ds + "分区数据为空，调度程序中止!");
                    System.exit(-1);
                } else {
                    log.info("company表{}分区数据不为空.", ds);
                }
                return;
            }
            verify(commandLine);
            String bizDate = commandLine.getOptionValue("b", DateUtils.getYesterday());
            String fileName = commandLine.getOptionValue("f");
            List<DataWorksFlowJob> jobs = SchemaInit.getJobs(fileName);

            if (commandLine.hasOption("s")) {
                log.info("执行单个业务流程！");
                String flow = commandLine.getOptionValue("flow");
                String task = commandLine.getOptionValue("task");
                SingleJobMain singleJobMain = context.getBean(SingleJobMain.class);
                String msg = "开始执行单个业务流程！bizDate:" + bizDate + " flow:" + flow + " task:" + task + " JobFile:" + commandLine.getOptionValue("f");
                log.info(msg);
                if (commandLine.hasOption("d")) {
                    log.info("debug模式：{}", jobs);
                } else {
                    singleJobMain.start(flow, task, bizDate, jobs);
                }
            } else {
                log.info("触发全部业务流程！");
                Main bean = context.getBean(Main.class);
                String msg = "开始执行全部业务流程！bizDate:" + bizDate + " JobFile:" + commandLine.getOptionValue("f");
                log.info(msg);
                if (commandLine.hasOption("d")) {
                    log.info("debug模式：{}", jobs);
                } else {
                    dd.send(msg);
                    bean.start(bizDate, jobs);
                }
            }
        } catch (ParseException e) {
            HelpFormatter formatter = new HelpFormatter();
            formatter.printHelp("很棒，参数错误", options);
            System.exit(-1);
        } catch (FileNotFoundException e) {
            log.error(e.getMessage());
            System.exit(-2);
        } catch (Exception e) {
            dd.send(e.getMessage());
            log.error(e.getMessage(), e);
            System.exit(-3);
        }
    }

    @SneakyThrows
    private void start(String bizDate, List<DataWorksFlowJob> jobs) {
        log.info("start!");
        LocalDateTime start = LocalDateTime.now();
        List<TaskInfo> collect = jobs.stream()
                .flatMap(dataWorksFlowJob -> {
                    List<DataWorksFlowTask> task = dataWorksFlowJob.getTask();
                    return task
                            .stream()
                            .map(dataWorksFlowTask -> {
                                TaskParam build = TaskParam.builder()
                                        .projectName(dataWorksFlowJob.getProject())
                                        .bizDate(bizDate)
                                        .flowName(dataWorksFlowJob.getFlow())
                                        .nodeParam(dataWorksFlowTask.toNodeParam(bizDate))
                                        .build();
                                CreateManualDagResponse r = touchService.touch(build);

                                return TaskInfo.builder().project(dataWorksFlowJob.getProject())
                                        .flow(dataWorksFlowJob.getFlow())
                                        .taskName(dataWorksFlowTask.getTaskName())
                                        .dagId(r.getReturnValue()).build();
                            });
                }).collect(Collectors.toList());

        int totalTask = collect.size();

        Set<TaskInfo> end = new HashSet<>();
        TimedCache<TaskInfo, String> timedCache = CacheUtil.newTimedCache(300 * 1000);
        int i = 0;
        int successTask = 0;
        int failedTask = 0;
        while (true) {
            int awaitTask = 0;

            boolean flag = true;
            List<String> empty = new ArrayList<>();

            for (TaskInfo taskInfo : collect) {
                if (end.contains(taskInfo)) {
                    continue;
                }
                Map<String, TaskFlowEnum> query = touchService.query(taskInfo.getProject(), taskInfo.getDagId());
                Map<String, List<String>> status = query.entrySet().stream()
                        .map(e -> {
                            Entry<String, String> entry = new Entry<>();
                            if (e.getValue().equals(TaskFlowEnum.SUCCESS) || e.getValue().equals(TaskFlowEnum.FAILURE)) {
                                entry.setKey(e.getValue().getMsg());
                            } else {
                                entry.setKey("运行中");
                            }
                            entry.setValue(e.getKey());
                            return entry;
                        }).collect(Collectors.groupingBy(Entry::getKey, Collectors.mapping(Entry::getValue, Collectors.toList())));
                List<String> success = status.getOrDefault(TaskFlowEnum.SUCCESS.getMsg(), empty);
                List<String> failure = status.getOrDefault(TaskFlowEnum.FAILURE.getMsg(), empty);
                List<String> await = status.getOrDefault("运行中", empty);


                if (failure.size() != 0) {
                    failedTask++;
                    log.error("failure node:{} ", failure);
                    DingMsg error = new DingMsg("任务失败", taskInfo.getProject(), taskInfo.getFlow(), String.join(",", failure), TaskFlowEnum.FAILURE.getMsg());
                    dingUtils.send(error);
                } else {
                    if (await.size() != 0) {
                        awaitTask++;
                    } else {
                        successTask++;
                    }
                }


                flag = flag && (await.size() == 0 || failure.size() != 0);

                if (await.size() == 0 || failure.size() != 0) {
                    end.add(taskInfo);
                } else {
                    LocalDateTime now = LocalDateTime.now();
                    Duration duration = Duration.between(start, now);
                    long l = duration.toHours();
                    if (l >= 2) {
                        if (!timedCache.containsKey(taskInfo) && i <= 6) {
                            //超两小时
                            i++;
                            DingMsg error = new DingMsg("【" + i + "】任务长时间未结束", taskInfo.getProject(), taskInfo.getFlow(), String.join(",", failure), TaskFlowEnum.RUNNING.getMsg());
                            dingUtils.send(error);
                            timedCache.put(taskInfo, "1");
                        }
                    }
                }
            }

            log.info("\nawait task:{}\ntotal task:{}\nsuccess task:{}\nfailure task:{}\n"
                    , awaitTask
                    , totalTask
                    , successTask
                    , failedTask
            );

            if (flag) {
                break;
            }
            Thread.sleep(10000);
        }
        if (failedTask != 0) {
            System.exit(-1);
        }
        log.info("end");
    }
}
